/*******************************************************************************
 * (c) Copyright 2011-2013 Microsemi SoC Products Group.  All rights reserved.
 *
 * SmartFusion2 Microcontroller Subsystem 10/100/1000 Mbps Ethernet MAC bare
 * metal software driver implementation.
 *
 * SVN $Revision: 5686 $
 * SVN $Date: 2013-04-12 13:24:08 +0100 (Fri, 12 Apr 2013) $
 */
#include <stdio.h>
#include "../../CMSIS/m2sxxx.h"
#include "../../CMSIS/mss_assert.h"
#include "../../CMSIS/hw_reg_io.h"
#include "mss_ethernet_mac_regs.h"
#include "mss_ethernet_mac.h"
#include "phy.h"
#include "mss_ethernet_mac_user_config.h"

#ifdef __cplusplus
extern "C" {
#endif

/**************************************************************************/
/* Preprocessor Macros                                                    */
/**************************************************************************/
#define SET       ENABLE
#define RESET     DISABLE

#define IS_STATE(x)             ( ((x) == MSS_MAC_ENABLE) || ((x) == MSS_MAC_DISABLE) )
#define IS_ENABLE(x)            ( (x) == MSS_MAC_ENABLE )
#define IS_DISABLE(x)           ( (x) == MSS_MAC_DISABLE )

#define IS_BYTE_MODE(x)         ( ((x) == MSS_MAC_BYTE_MODE ) )
#define IS_NIBBLE_MODE(x)       ( ((x) == MSS_MAC_NIBBLE_MODE) )

#define GET_DESC_EMPTY_FLAG(x)          ( (x) & DMA_DESC_EMPTY_FLAG_MASK )
#define SET_DESC_EMPTY_FLAG(x)          ( (x) = (x) | (DMA_DESC_EMPTY_FLAG_MASK) )
#define CLEAR_DESC_EMPTY_FLAG(x)        ( (x) = (x) & (uint32_t)(~DMA_DESC_EMPTY_FLAG_MASK) )
#define GET_PKT_SIZE(x)                 ( (x) & DMA_DESC_PKT_SIZE_MASK )

#define MAC_STAT_REG_OFFSET             ( 0x00000080u )
#define MAC_STAT_BASE                   ( ETHERNET_BASE + MAC_STAT_REG_OFFSET )
#define MAC_STAT_REG(x)                 ( (uint32_t *)( MAC_STAT_BASE + ((x) << 2u)) )

#define IS_WORD_ALIGNED(x)      ((uint32_t)0 == ((uint32_t)x & (uint32_t)3))

/*******************************************************************************
 * DMA Descriptor
 * Driver creates and manages two descriptor rings for transmission and reception.
 */
typedef struct
{
    uint32_t pkt_start_addr;    /* Packet start address */
    volatile uint32_t pkt_size; /* Packet size & Per packet override flags */
    uint32_t next_desriptor;    /* Link to next descriptor */
    uint32_t index;             /* Index: helps in handling interrupts */
    void * caller_info;         /* Pointer to user specific data */
} MAC_desc_t;

/*******************************************************************************
 * SmartFusion2 Ethernet MAC instance
 * A local record of this type "g_mac" will be created and maintained by the
 * driver.
 */
typedef struct
{
    MAC_TypeDef               * hw_reg;                            /* MAC registers base: Useful for debugging */
    IRQn_Type                   irqn;                              /* IRQ number assigned to MAC */
    MAC_cfg_t                 * cfg;                               /* Pointer to current configuration record */
    MAC_desc_t                  tx_desc_tab[MSS_MAC_TX_RING_SIZE]; /* Transmit descriptor table */
    MAC_desc_t                * curr_tx_desc;                      /* Pointer to current transmit descriptor in use */
    MAC_desc_t                * next_tx_desc;                      /* Pointer to next free transmit descriptor */
    MAC_desc_t                  rx_desc_tab[MSS_MAC_RX_RING_SIZE]; /* Receive descriptor table */
    MAC_desc_t                * curr_rx_desc;                      /* Pointer to current receive descriptor in use */
    MAC_desc_t                * next_rx_desc;                      /* Pointer to next free receive descriptor */
    uint8_t                     mac_addr[6];                       /* Station's MAC address */
    MAC_stat_t                  stat;                              /* Latest statistics read / cleared */
    MAC_transmit_callback_t     tx_complete_handler;
    MAC_receive_callback_t      pckt_rx_callback;
    int16_t                     first_tx_index;
    int16_t                     last_tx_index;
} MAC_instance_t;

/*******************************************************************************
 * MAC interrupt definitions
 * This type definition provides interrupt indices for MAC hardware.
 */
typedef enum
{
    MSS_MAC_TXPKTSENT_IRQ = 0u,
    MSS_MAC_TXUNDRRUN_IRQ = 1u,
    MSS_MAC_TXBUSERR_IRQ  = 3u,
    MSS_MAC_RXPKTRCVD_IRQ = 4u,
    MSS_MAC_RXOVRFLOW_IRQ = 6u,
    MSS_MAC_RXBUSERR_IRQ  = 7u,
    MSS_MAC_INVALID_IRQ   = 255u
} mac_irq_t;

/*******************************************************************************
 * MAC I/O direction.
 * This type definition provides INPUT or OUTPUT direction of communication.
 */
typedef enum
{
    RX,
    TX,
} MAC_io_t;

/**************************************************************************/
/* Type defines                                                           */
/**************************************************************************/

/**************************************************************************/
/* Global variables                                                       */
/**************************************************************************/

static MAC_instance_t g_mac;

/*******************************************************************************
 * Default MAC address 
 */
static const uint8_t g_default_mac_address[6] =
{
    0xC0u, 0xB1u, 0x3Cu, 0x88u, 0x88u, 0x88u
};

/**************************************************************************/
/* Private Functions                                                      */
/**************************************************************************/
static void mac_reset(void);
static void desc_ring_init(MAC_desc_t * ring, MAC_io_t io);
static uint8_t * mac_memcpy(uint8_t *dest, const uint8_t *src, uint32_t n);
static void assign_station_addr(const uint8_t *mac_addr);
static void rxpkt_handler(void);
static void txpkt_handler(void);
static void update_mac_cfg(void);

/**************************************************************************/
/* Public Functions                                                       */
/**************************************************************************/

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void 
MSS_MAC_init
( 
    MAC_cfg_t * cfg 
)
{
    uint32_t tempreg = 0;

    ASSERT(cfg != NULL);

    if(cfg != NULL)
    {
        NVIC_DisableIRQ(EthernetMAC_IRQn);
        mac_reset();
        NVIC_ClearPendingIRQ(EthernetMAC_IRQn);
        
        /* Check for validity of configuration parameters */
        ASSERT( INVALID_INF > cfg->interface );
        ASSERT( IS_STATE(cfg->tx_edc_enable) );
        ASSERT( IS_STATE(cfg->rx_edc_enable) );        
        ASSERT( MSS_MAC_PREAMLEN_MAXVAL >= cfg->preamble_length );
        ASSERT( IS_STATE(cfg->hugeframe_enable) );
        ASSERT( IS_STATE(cfg->length_field_check) );
        ASSERT( IS_STATE(cfg->pad_n_CRC) );
        ASSERT( IS_STATE(cfg->append_CRC) );
        ASSERT( IS_STATE(cfg->loopback) );
        ASSERT( IS_STATE(cfg->rx_flow_ctrl) );
        ASSERT( IS_STATE(cfg->tx_flow_ctrl) );
        ASSERT( MSS_MAC_MINIFG_MAXVAL >= cfg->min_IFG );            
        ASSERT( MSS_MAC_BTBIFG_MAXVAL >= cfg->btb_IFG );           
        ASSERT( MSS_MAC_MAXRETX_MAXVAL >= cfg->max_retx_tries );      
        ASSERT( IS_STATE(cfg->excessive_defer) );    
        ASSERT( IS_STATE(cfg->nobackoff) );         
        ASSERT( IS_STATE(cfg->backpres_nobackoff) );
        ASSERT( IS_STATE(cfg->ABEB_enable) );        
        ASSERT( MSS_MAC_ABEBTRUNC_MAXVAL >= cfg->ABEB_truncvalue );    
        ASSERT( MSS_MAC_BY28_PHY_CLK >= cfg->phyclk );            
        ASSERT( IS_STATE(cfg->supress_preamble) );   
        ASSERT( IS_STATE(cfg->autoscan_phys) );      
        ASSERT( MSS_MAC_MAXFRAMELEN_MAXVAL >= cfg->max_frame_length );   
        ASSERT( MSS_MAC_SLOTTIME_MAXVAL >= cfg->slottime );   

        g_mac.first_tx_index = -1;
        g_mac.last_tx_index = -1;
        
        /* Tx FIFO EDAC configuration */
        ( IS_ENABLE(cfg->tx_edc_enable) ? set_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_TX_EN) :
                                          clear_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_TX_EN) );
        /* Rx FIFO EDAC configuration */
        ( IS_ENABLE(cfg->rx_edc_enable) ? set_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_RX_EN) :
                                          clear_bit_reg32(&SYSREG->EDAC_CR,MAC_EDAC_RX_EN) );

        /* Configure PHY related MII MGMT registers */
        tempreg = (uint32_t)cfg->phyclk;
        ( IS_ENABLE(cfg->supress_preamble) ? (tempreg |= MII_PREAM_SUPRESS_MASK):
                                             (tempreg &= ~MII_PREAM_SUPRESS_MASK) );
        ( IS_ENABLE(cfg->autoscan_phys) ? (tempreg |= MII_SCAN_AUTO_INC_MASK):
                                             (tempreg &= ~MII_SCAN_AUTO_INC_MASK) );  
        MAC->MII_CONFIG = tempreg;
        
        /* Initialize MAC instance */
        g_mac.hw_reg = MAC;
        g_mac.irqn   = EthernetMAC_IRQn;
        g_mac.cfg    = cfg;
        g_mac.stat   = MAC_LAST_STAT;

        /* Intialize Tx & Rx descriptor rings */
        desc_ring_init( &g_mac.tx_desc_tab[0], TX);
        desc_ring_init( &g_mac.rx_desc_tab[0], RX);
        
        /* 
         * At the time of intialization next descriptors
         * and Current descriptors will be NULL as no tx/rx is performed.
         */
        g_mac.curr_tx_desc = NULL;
        g_mac.next_tx_desc = &g_mac.tx_desc_tab[0];
        g_mac.curr_rx_desc = NULL;
        g_mac.next_rx_desc = &g_mac.rx_desc_tab[0];   
        /*
         * MAC hardware configuration
         */              
        /* Clear soft reset to keep MCXMAC functional */
        clear_bit_reg32( &MAC->CFG1, CFG1_SOFT_RST);

        /* Clear all reset bits */
        /* Clear MCXMAC TX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_TX_RST);
        /* Clear MCXMAC RX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_RX_RST);
        /* Clear MCXMAC TX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_TXCTL_RST);
        /* Clear MCXMAC RX reset */
        clear_bit_reg32( &MAC->CFG1, CFG1_RXCTL_RST);
        /* Clear MCXMAC interface reset */
        clear_bit_reg32( &MAC->INTERFACE_CTRL, MII_INTF_RESET);
        /* Clear FIFO watermark reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_WMM_RST);
        /* Clear FIFO Rx system reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RSYS_RST);
        /* Clear FIFO Rx fab reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RFAB_RST);
        /* Clear FIFO Tx system reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TSYS_RST);
        /* Clear FIFO Tx system reset */
        clear_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TFAB_RST);

        ( IS_ENABLE(cfg->loopback) ? ( tempreg = CFG1_LOOPBACK_MASK) : (tempreg = 0) );
        ( IS_ENABLE(cfg->rx_flow_ctrl) ? ( tempreg |= CFG1_RX_FCTL_MASK) :
                                         ( tempreg &= ~CFG1_RX_FCTL_MASK));                                     
        ( IS_ENABLE(cfg->tx_flow_ctrl) ? ( tempreg |= CFG1_TX_FCTL_MASK) :
                                         ( tempreg &= ~CFG1_TX_FCTL_MASK));
        MAC->CFG1 = tempreg ;

        ( IS_ENABLE(cfg->fullduplex) ? ( tempreg = CFG2_FDX_MASK) : (tempreg = 0) );
        ( IS_ENABLE(cfg->append_CRC) ? ( tempreg |= CFG2_CRC_EN_MASK) :
                                       ( tempreg &= ~CFG2_CRC_EN_MASK));
        ( IS_ENABLE(cfg->pad_n_CRC) ? ( tempreg |= CFG2_PAD_CRC_EN_MASK) :
                                      ( tempreg &= ~CFG2_PAD_CRC_EN_MASK));
        ( IS_ENABLE(cfg->length_field_check) ? ( tempreg |= CFG2_LEN_CHECK_MASK) :
                                               ( tempreg &= ~CFG2_LEN_CHECK_MASK));        
        ( IS_ENABLE(cfg->hugeframe_enable) ? ( tempreg |= CFG2_HUGE_FRAME_EN_MASK) :
                                             ( tempreg &= ~CFG2_HUGE_FRAME_EN_MASK));
                                             
        if((MII == cfg->interface) || (RMII == cfg->interface))
        {
            /* MII and RMII use nibble mode interface. */
            tempreg |= CFG2_NIBBLE_MASK;
        }
        else
        {
            /* TBI and GMII use byte interface. */
            tempreg |= CFG2_BYTE_MASK;
        }
        tempreg |= ( ((uint32_t)cfg->preamble_length ) << CFG2_PREAM_LEN );
        MAC->CFG2 = tempreg ;
        
        tempreg = cfg->btb_IFG;
        tempreg |= ( (uint32_t)cfg->min_IFG << IFG_MINIFGENF );
        tempreg |= ( cfg->non_btb_IFG << IFG_NONBTBIPG );
        MAC->IFG = tempreg;
        
        tempreg = cfg->slottime;
        tempreg |= ( (uint32_t)cfg->max_retx_tries << HALF_DUPLEX_RETX_MAX);
        ( IS_ENABLE(cfg->excessive_defer)? (tempreg |= HALF_DUPLEX_EXCS_DEFER_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_EXCS_DEFER_MASK) ); 
        ( IS_ENABLE(cfg->nobackoff)? (tempreg |= HALF_DUPLEX_NO_BACKOFF_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_NO_BACKOFF_MASK) );
        ( IS_ENABLE(cfg->backpres_nobackoff)? (tempreg |= HALF_DUPLEX_BACKPRES_NOBACKOFF_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_BACKPRES_NOBACKOFF_MASK) );
        ( IS_ENABLE(cfg->ABEB_enable)? (tempreg |= HALF_DUPLEX_ABEB_ENABLE_MASK) :
                                           (tempreg &= ~HALF_DUPLEX_ABEB_ENABLE_MASK) );       
        tempreg |= ( (uint32_t)cfg->ABEB_truncvalue << HALF_DUPLEX_ABEB_TUNC );
        MAC->HALF_DUPLEX = tempreg;

        MAC->MAX_FRAME_LENGTH = (uint32_t)cfg->max_frame_length;

        /* Configure FIFO */
        MAC->FIFO_CFG0 = ( FIFO_CFG0_WMM_ENABLE  | FIFO_CFG0_RSYS_ENABLE |
                           FIFO_CFG0_RFAB_ENABLE | FIFO_CFG0_TSYS_ENABLE |
                           FIFO_CFG0_TFAB_ENABLE );
        /* RX FIFO size : 8KB  */        
        MAC->FIFO_CFG1 = FIFO_CFG1_DEFVAL;
        /* Rx FIFO watermark */
        MAC->FIFO_CFG2 = FIFO_CFG2_DEFVAL;
        /* Tx FIFO watermark: 4KB Tx FIFO */
        MAC->FIFO_CFG3 = FIFO_CFG3_DEFVAL;
        /* Drop junk frames (Frames suffered CRC error or short collision) */
        MAC->FIFO_CFG4 = FIFO_CFG4_DEFVAL;
        /* Drop frames with CRC errors and PAUSE frames */
        MAC->FIFO_CFG5 = FIFO_CFG5_DEFVAL;

        /* Assign MAC station address */
        assign_station_addr(cfg->mac_addr);

        /* initialize default interrupt handlers */
        g_mac.tx_complete_handler = NULL;
        g_mac.pckt_rx_callback = NULL;
        
        /* Interface type: MII/GMII/RMII/TBI */
        SYSREG->MAC_CR = ((uint32_t)cfg->interface) << MAC_CONFIG_INTF;

        /* Initialize PHY interface */
        phy_init(cfg);
        phy_set_link_speed(cfg->speed_duplex_select);
        phy_autonegotiate();
        update_mac_cfg();
        
        /* Enable transmission at MAC level. */
        set_bit_reg32( &MAC->CFG1, CFG1_TX_EN);
        /* Enable reception at MAC level.    */
        set_bit_reg32( &MAC->CFG1, CFG1_RX_EN);
        
        /* Enable Tx Packet interrupt */
        set_bit_reg32(&MAC->DMA_IRQ_MASK, MSS_MAC_TXPKTSENT_IRQ);
        /* Enable RX Packet interrupt */
        set_bit_reg32(&MAC->DMA_IRQ_MASK, MSS_MAC_RXPKTRCVD_IRQ);
    }
}

/**************************************************************************//**
 * 
 */
static void update_mac_cfg(void)
{
    MAC_speed_t speed;
    uint8_t fullduplex;
    uint8_t link_up;
    
    link_up = phy_get_link_status(&speed, &fullduplex);
    if(link_up != 0)
    {
        /* Reconfigure MAC based on PHY configuration. */
        if(MAC100MBPS == speed)
        {
            set_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
        }
        else
        {
            clear_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
        }
        
        SYSREG->MAC_CR = (SYSREG->MAC_CR & ~MAC_CONFIG_SPEED_MASK) | speed;
        
        /* Configure duplex mode */
        if(0u == fullduplex)
        {
            /* half duplex */
            MAC->CFG2 &= ~CFG2_FDX_MASK;
        }
        else
        {
            /* full duplex */
            MAC->CFG2 |= CFG2_FDX_MASK;
        }
    }
}

uint8_t MSS_MAC_get_link_status
(
    MAC_speed_t * speed,
    uint8_t *     fullduplex
)
{
    static MAC_speed_t previous_speed = INVALID_SPEED;
    static uint8_t previous_duplex = 0xAA;
    MAC_speed_t link_speed;
    uint8_t link_fullduplex;
    uint8_t link_up;
    
    link_up = phy_get_link_status(&link_speed, &link_fullduplex);
    if(link_up != 0)
    {
        /*----------------------------------------------------------------------
         * Update MAC configuration if link characteristics changed.
         */
        if(link_speed != previous_speed)
        {
            /* Reconfigure MAC based on PHY configuration. */
            if(MAC100MBPS == link_speed)
            {
                set_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
            }
            else
            {
                clear_bit_reg32(&MAC->INTERFACE_CTRL,INF_SPEED);
            }
            
            SYSREG->MAC_CR = (SYSREG->MAC_CR & ~MAC_CONFIG_SPEED_MASK) | link_speed;
        }
        previous_speed = link_speed;
        
        if(link_fullduplex != previous_duplex)
        {
            /* Configure duplex mode */
            if(0u == link_fullduplex)
            {
                /* half duplex */
                MAC->CFG2 &= ~CFG2_FDX_MASK;
            }
            else
            {
                /* full duplex */
                MAC->CFG2 |= CFG2_FDX_MASK;
            }
        }
        previous_duplex = link_fullduplex;
        
        /*----------------------------------------------------------------------
         * Return current link speed and duplex mode.
         */
        if(speed != NULL)
        {
            *speed = link_speed;
        }
        
        if(fullduplex != NULL)
        {
            *fullduplex = link_fullduplex;
        }
    }
    
    return link_up;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void 
MSS_MAC_cfg_struct_def_init
(
    MAC_cfg_t * cfg
)
{
    ASSERT( NULL != cfg);
    if(NULL != cfg)
    {
        cfg->speed_duplex_select = MSS_MAC_ANEG_ALL_SPEEDS;
        cfg->phy_addr           = 0u;
        cfg->interface          = MII;
        cfg->tx_edc_enable      = MSS_MAC_ERR_DET_CORR_ENABLE;
        cfg->rx_edc_enable      = MSS_MAC_ERR_DET_CORR_ENABLE;
        cfg->preamble_length    = MSS_MAC_PREAMLEN_DEFVAL;
        cfg->hugeframe_enable   = MSS_MAC_HUGE_FRAME_DISABLE;
        cfg->length_field_check = MSS_MAC_LENGTH_FILED_CHECK_ENABLE;
        cfg->pad_n_CRC          = MSS_MAC_PAD_N_CRC_ENABLE;
        cfg->append_CRC         = MSS_MAC_CRC_ENABLE;
        cfg->fullduplex         = MSS_MAC_FULLDUPLEX_ENABLE;
        cfg->loopback           = MSS_MAC_LOOPBACK_DISABLE;
        cfg->rx_flow_ctrl       = MSS_MAC_RX_FLOW_CTRL_DISABLE;
        cfg->tx_flow_ctrl       = MSS_MAC_TX_FLOW_CTRL_DISABLE;
        cfg->min_IFG            = MSS_MAC_MINIFG_DEFVAL;
        cfg->btb_IFG            = MSS_MAC_BTBIFG_DEFVAL;
        cfg->max_retx_tries     = MSS_MAC_MAXRETX_DEFVAL;
        cfg->excessive_defer    = MSS_MAC_EXSS_DEFER_DISABLE;
        cfg->nobackoff          = MSS_MAC_NO_BACKOFF_DISABLE;
        cfg->backpres_nobackoff = MSS_MAC_BACKPRESS_NO_BACKOFF_DISABLE;
        cfg->ABEB_enable        = MSS_MAC_ABEB_DISABLE;
        cfg->ABEB_truncvalue    = MSS_MAC_ABEBTRUNC_DEFVAL;
        cfg->phyclk             = MSS_MAC_DEF_PHY_CLK;
        cfg->supress_preamble   = MSS_MAC_SUPPRESS_PREAMBLE_DISABLE;
        cfg->autoscan_phys      = MSS_MAC_PHY_AUTOSCAN_DISABLE;
        cfg->max_frame_length   = MSS_MAC_MAXFRAMELEN_DEFVAL;
        cfg->non_btb_IFG        = MSS_MAC_NONBTBIFG_DEFVAL;
        cfg->slottime           = MSS_MAC_SLOTTIME_DEFVAL;
        cfg->framedrop_mask     = MSS_MAC_FRAME_DROP_MASK_DEFVAL;
        
        mac_memcpy(cfg->mac_addr, g_default_mac_address, 6u);
    }
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
static void 
mac_reset
(
    void
)
{
    /* Reset MCXMAC TX functionality */
    set_bit_reg32( &MAC->CFG1, CFG1_TX_RST);
    /* Reset MCXMAC RX functionality */
    set_bit_reg32( &MAC->CFG1, CFG1_RX_RST);
    /* Reset MCXMAC TX control */
    set_bit_reg32( &MAC->CFG1, CFG1_TXCTL_RST);
    /* Reset MCXMAC RX control */
    set_bit_reg32( &MAC->CFG1, CFG1_RXCTL_RST);
    /* Reset MCXMAC interface control */
    set_bit_reg32( &MAC->INTERFACE_CTRL, MII_INTF_RESET);
    /* Reset FIFO watermark module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_WMM_RST);
    /* Reset FIFO Rx system module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RSYS_RST);
    /* Reset FIFO Rx fab module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_RFAB_RST);
    /* Reset FIFO Tx system module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TSYS_RST);
    /* Reset FIFO Tx system module */
    set_bit_reg32( &MAC->FIFO_CFG0, FIFO_CFG0_TFAB_RST);
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
MSS_MAC_write_phy_reg
(
    uint8_t phyaddr,
    uint8_t regaddr,
    uint16_t regval
)
{
    ASSERT( MSS_MAC_PHYADDR_MAXVAL >= phyaddr); 
    ASSERT( MSS_MAC_PHYREGADDR_MAXVAL >= regaddr );
    /* 
     * Write PHY address in MII Mgmt address register.
     * Makes previous register address 0 & invalid.
     */ 
    if((MSS_MAC_PHYADDR_MAXVAL >= phyaddr) && 
       (MSS_MAC_PHYREGADDR_MAXVAL >= regaddr))
    {
        /* Wait for MII Mgmt interface to complete previous operation. */
        do {
            ;
        } while(MAC->MII_INDICATORS & MII_BUSY_INDICATOR_MASK);
    
        /* Load PHY address in MII Mgmt address register */
        MAC->MII_ADDRESS = (((uint32_t)phyaddr)<<MII_PHY_ADDR);
        /* Load register address in MII Mgmt address register */
        MAC->MII_ADDRESS |= (uint32_t)regaddr;
        /* Load value to be written in MII Mgmt control register */
        MAC->MII_CTRL = (uint32_t)regval ;
    }
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint16_t
MSS_MAC_read_phy_reg
(
    uint8_t phyaddr,
    uint8_t regaddr
)
{
    uint32_t timeout = 100000u;
    uint16_t read_val = 0;

    ASSERT( MSS_MAC_PHYADDR_MAXVAL >= phyaddr); 
    ASSERT( MSS_MAC_PHYREGADDR_MAXVAL >= regaddr );
    /* 
     * Write PHY address in MII Mgmt address register.
     * Makes previous register address 0 & invalid.
     */ 
    if( ( MSS_MAC_PHYADDR_MAXVAL >= phyaddr) && 
        ( MSS_MAC_PHYREGADDR_MAXVAL >= regaddr ) )
    {
        /* Wait for MII Mgmt interface to complete previous operation. */
        do {
            ;
        } while(MAC->MII_INDICATORS & MII_BUSY_INDICATOR_MASK);
        
        MAC->MII_ADDRESS = (((uint32_t)phyaddr) << MII_PHY_ADDR);
        /* Load PHY register address in MII Mgmt address register */
        MAC->MII_ADDRESS |= (uint32_t)regaddr;
        /* Issue read command to PHY */
        set_bit_reg32( &MAC->MII_COMMAND, MII_READ);
        /* Poll till PHY read cycle is completed or timeout expired */
        while ( timeout-- && ((MAC->MII_INDICATORS & MII_BUSY_INDICATOR_MASK) || 
                              (MAC->MII_INDICATORS & MII_NV_INDICATOR_MASK)) );
        /* Read value from MII Mgmt status register only if timeout is not expired
           but the read the read cycle is completed */
        if( timeout != 0 )
            read_val = (uint16_t)MAC->MII_STATUS ;
    }
    clear_bit_reg32( &MAC->MII_COMMAND, MII_READ);
    return read_val;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint32_t 
MSS_MAC_read_stat
(
    MAC_stat_t stat
)
{
    uint32_t stat_val = 0;

    ASSERT ( MAC_LAST_STAT > stat );
    if( MAC_LAST_STAT > stat )
        stat_val = *(MAC_STAT_REG(stat));

    return stat_val;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t
MSS_MAC_receive_pkt
(
    uint8_t * rx_pkt_buffer,
    void * caller_info
)
{
    uint8_t status = MSS_MAC_FAILED;

    /* Make this function atomic w.r.to EMAC interrupt */
    NVIC_DisableIRQ(EthernetMAC_IRQn);
    
    ASSERT(NULL != rx_pkt_buffer);
    ASSERT(IS_WORD_ALIGNED(rx_pkt_buffer));

    if((NULL != rx_pkt_buffer) && (IS_WORD_ALIGNED(rx_pkt_buffer)))
    {
        if((NULL != g_mac.next_rx_desc) && 
           (!GET_DESC_EMPTY_FLAG(g_mac.next_rx_desc->pkt_size)))
        {
            g_mac.next_rx_desc->pkt_start_addr = (uint32_t)rx_pkt_buffer;
            g_mac.next_rx_desc->caller_info = caller_info;

            /* Set own flag of this descriptor to indicate ready for RX */
            SET_DESC_EMPTY_FLAG(g_mac.next_rx_desc->pkt_size);

            /* 
               If the RX is found disabled, then it might be because this is the
               first time a packet is scheduled for reception or the RX ENABLE is 
               made zero by RX overflow or RX bus error. In either case, this
               function tries to schedule the current packet for reception.
            */
            if(!read_bit_reg32(&MAC->DMA_RX_CTRL, DMA_RXENABLE))
            {
                /* Make current descriptor to point to the packet requested */
                g_mac.curr_rx_desc = g_mac.next_rx_desc;
                MAC->DMA_RX_DESC = (uint32_t)g_mac.curr_rx_desc;
                set_bit_reg32(&MAC->DMA_RX_STATUS, DMA_RXOVRFLOW);
                set_bit_reg32(&MAC->DMA_RX_CTRL, DMA_RXENABLE);
            }

            /* Point the next_rx_desc to next free descriptor in the ring */
            /* Wrap around in case next descriptor is pointing to last in the ring */
            if((MSS_MAC_RX_RING_SIZE - 1u) == g_mac.next_rx_desc->index)
            {
                g_mac.next_rx_desc = &g_mac.rx_desc_tab[0];
            }
            else
            {
                ++g_mac.next_rx_desc;
            }
                
            /* Packet scheduled for reception successfully only when there 
               is no new RX bus error */
            if(!read_bit_reg32(&MAC->DMA_RX_STATUS, DMA_RXBUSERR))
            {
                status = MSS_MAC_SUCCESS;
            }
        }
    }
    /* Ethernet Interrupt Enable function. */
    NVIC_EnableIRQ(EthernetMAC_IRQn);
    return status;
}

/***************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
uint8_t 
MSS_MAC_send_pkt
(
    uint8_t * tx_buffer,
    uint32_t tx_length,
    void * caller_info
)
{
    uint8_t status = MSS_MAC_FAILED;

    /* Make this function atomic w.r.to EMAC interrupt */
    NVIC_DisableIRQ(EthernetMAC_IRQn);

    ASSERT(NULL != tx_buffer);
    ASSERT(0 != tx_length);
    ASSERT(IS_WORD_ALIGNED(tx_buffer));
    
    if((0 != tx_length) && (IS_WORD_ALIGNED(tx_buffer)))
    {
        if((NULL != g_mac.next_tx_desc) &&
           (GET_DESC_EMPTY_FLAG(g_mac.next_tx_desc->pkt_size)))
        {
            g_mac.next_tx_desc->pkt_start_addr = (uint32_t)tx_buffer;
            g_mac.next_tx_desc->pkt_size |= tx_length;
            g_mac.next_tx_desc->caller_info = caller_info;
            g_mac.last_tx_index = g_mac.next_tx_desc->index;
            if(-1 == g_mac.first_tx_index)
            {
                g_mac.first_tx_index = g_mac.next_tx_desc->index;
            }
            
            /* Make own flag 0 for this descriptor to indicate ready for tx */
            CLEAR_DESC_EMPTY_FLAG(g_mac.next_tx_desc->pkt_size);

            /*
                If TX is found disabled, this might be because this is
                the first time packet is being sent or the DMA completed 
                previous transfer and stopped transmission by itself caused
                by TX underrun or bus error. This function neglects the errors
                and tries to send the current packet.
             */
            if(!read_bit_reg32(&MAC->DMA_TX_CTRL, DMA_TXENABLE))
            {
                g_mac.curr_tx_desc = g_mac.next_tx_desc;
                MAC->DMA_TX_DESC = (uint32_t)g_mac.curr_tx_desc;
                set_bit_reg32( &MAC->DMA_TX_STATUS, DMA_TXUNDRRUN );
                set_bit_reg32( &MAC->DMA_TX_CTRL, DMA_TXENABLE );
            }

            /* Point the next_tx_desc to next free descriptor in the ring */
            /* Wrap around in case next descriptor is pointing to last in the ring */
            if((MSS_MAC_TX_RING_SIZE - 1u) == g_mac.next_tx_desc->index)
            {
                g_mac.next_tx_desc = &g_mac.tx_desc_tab[0];
            }
            else
            {
                ++g_mac.next_tx_desc;
            }
            /* Packet scheduled for transmission successfully only when 
               there is no TX bus error */
            if(!read_bit_reg32( &MAC->DMA_TX_STATUS, DMA_TXBUSERR))
            {
                status = MSS_MAC_SUCCESS;
            }
        }
    }
    
    /* Ethernet Interrupt Enable function. */
    NVIC_EnableIRQ(EthernetMAC_IRQn);
    return status;
}
/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void
EthernetMAC_IRQHandler
(
    void
)
{
    uint32_t dma_irq;
    
    dma_irq = MAC->DMA_IRQ;

    /* Transmit packet sent interrupt */
    if(dma_irq & (1u << MSS_MAC_TXPKTSENT_IRQ))
    {
        txpkt_handler();
    }
    /* Packet received interrupt */
    if(dma_irq & (1u << MSS_MAC_RXPKTRCVD_IRQ))
    {
        rxpkt_handler();
    }
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void MSS_MAC_set_tx_callback
(
    MAC_transmit_callback_t tx_complete_handler
)
{
    g_mac.tx_complete_handler = tx_complete_handler;
}

/**************************************************************************//**
 * See mss_ethernet_mac.h for details of how to use this function.
 */
void MSS_MAC_set_rx_callback
(
    MAC_receive_callback_t rx_callback
)
{
    g_mac.pckt_rx_callback = rx_callback;
}

/**************************************************************************/
/* Private Function definitions                                           */
/**************************************************************************/  

/**************************************************************************//**
 * This is default "Receive packet interrupt handler. This function finds the
 * descriptor that received the packet and caused the interrupt.
 * This informs the received packet size to the application and
 * relinquishes the packet buffer from the associated DMA descriptor.
 */
static void 
rxpkt_handler
(
    void
)
{
    uint32_t rxcnt = 0u;

    /* Execution comes here because at-least one packet is received.
       Hence rxcnt can be read after freeing one descriptor */
    do
    {
        MAC_desc_t * cdesc = g_mac.curr_rx_desc;

        /* Point the curr_rx_desc to next descriptor in the ring */
        /* Wrap around in case next descriptor is pointing to last in the ring */
        if((MSS_MAC_RX_RING_SIZE - 1u) == g_mac.curr_rx_desc->index)
        {
            g_mac.curr_rx_desc = &g_mac.rx_desc_tab[0];
        }
        else
        {
            ++g_mac.curr_rx_desc;
        }
        
        /* Clear the rx packet received interrupt once. If this 
           bit still persists, then another rx packet received 
           interrupt will be generated. Rx count will be decremented.*/
        set_bit_reg32( & MAC->DMA_RX_STATUS, DMA_RXPKTRCVD);
        /* Get the latest current received count */
        rxcnt = (MAC->DMA_RX_STATUS & DMA_PKTCOUNT_MASK) >> DMA_PKTCOUNT;
        
        /* Pass received packet up to application layer. */
        if(NULL != g_mac.pckt_rx_callback)
        {
            uint8_t * p_rx_packet;
            uint32_t pckt_length;
            
            p_rx_packet = (uint8_t *)cdesc->pkt_start_addr;
            pckt_length = (cdesc->pkt_size & DMA_DESC_PKT_SIZE_MASK) - 4u;
            if(NULL != g_mac.pckt_rx_callback)
            {
                g_mac.pckt_rx_callback(p_rx_packet, pckt_length, cdesc->caller_info);
            }
        }
    } while(0u != rxcnt);
}

/**************************************************************************//**
 * This is default "Transmit packet interrupt handler. This function finds the
 * descriptor that transmitted the packet and caused the interrupt.
 * This relinquishes the packet buffer from the associated DMA descriptor.
 */
static void 
txpkt_handler
(
    void
)
{
    uint32_t empty_flag;
    int16_t index;
    uint32_t completed = 0u;
        
    ASSERT(g_mac.first_tx_index != -1);
    ASSERT(g_mac.last_tx_index != -1);
    
    index = g_mac.first_tx_index;
    do
    {
        /* Call packet Tx completion handler if it exists. */
        if(NULL != g_mac.tx_complete_handler)
        {
            g_mac.tx_complete_handler(g_mac.tx_desc_tab[index].caller_info);
        }
        
        if(index == g_mac.last_tx_index)
        {
            /* all pending tx packets sent. */
            g_mac.first_tx_index = -1;
            g_mac.last_tx_index = -1;
            completed = 1u;
        }
        else
        {
            /* Move on to next transmit descriptor. */
            ++index;
            if(index >= MSS_MAC_TX_RING_SIZE)
            {
                index = 0;
            }
            g_mac.first_tx_index = index;
            
            /* Check if we reached a descriptor still pending tx. */
            empty_flag = GET_DESC_EMPTY_FLAG(g_mac.tx_desc_tab[index].pkt_size);
            if(0u == empty_flag)
            {
                completed = 1u;
            }
        }
        
    } while(0u == completed);
    
    /* Clear the tx packet sent interrupt. */
    set_bit_reg32(&MAC->DMA_TX_STATUS, DMA_TXPKTSENT);
}

/**************************************************************************//**
 * 
 */                                  
static void 
desc_ring_init
(
    MAC_desc_t * ring, 
    MAC_io_t io
)
{
    uint32_t l = 0u;
    uint32_t size = 0u;

    (TX == io) ? ( size = MSS_MAC_TX_RING_SIZE) : (size = MSS_MAC_RX_RING_SIZE);
    for(l=0; l<size; l++)    
    {
        /* Set packet start address to NULL */
        ring[l].pkt_start_addr = (uint32_t)NULL;
        /* 
         * Set empty flag indicating invalid data in case of Tx descriptor.
         * clear empty flag indicating full buffer in case of Rx descriptor.
         * This disables the DMA from transferring packets to and from FIFO 
         * in case of Tx descriptor and Rx descriptors respectively.
         */
        (TX == io) ? SET_DESC_EMPTY_FLAG(ring[l].pkt_size) : 
                            CLEAR_DESC_EMPTY_FLAG(ring[l].pkt_size);
        /* Every descriptor links to next descriptor */
        ring[l].next_desriptor = (uint32_t)(&ring[l+1]);
//<CJ>        ring[l].pkt = NULL;
        ring[l].index = l;
    }
    /* Complete the ring by connecting tail and head */
    ring[l-1u].next_desriptor = (uint32_t)(&ring[0]);
    
}

/**************************************************************************//**
 * 
 */
static void assign_station_addr
(
    const uint8_t *mac_addr
)
{
    uint32_t address32;

    ASSERT(NULL!= mac_addr);

    if(NULL != mac_addr)
    {
        address32 = ((uint32_t)mac_addr[0]) << 24u;
        address32 |= ((uint32_t)mac_addr[1]) << 16u;
        address32 |= ((uint32_t)mac_addr[2]) << 8u;
        address32 |= ((uint32_t)mac_addr[3]);
        MAC->STATION_ADDRESS1 = address32;
        address32 = ((uint32_t)mac_addr[4]) << 24u;
        address32 |= ((uint32_t)mac_addr[5]) << 16u;
        MAC->STATION_ADDRESS2 = address32;
    }
}

/***************************************************************************//**
 * MAC_memcpy()
 * This copies n bytes from memory area src to memory area dest.
 * The memory areas should not overlap.
 *
 * @return          a pointer to the memory area dest.
 */
static uint8_t * mac_memcpy(uint8_t *dest, const uint8_t *src, uint32_t n)
{
    uint8_t *d = dest;

    while( n > 0 ) {
        n--;
        d[n] = src[n];
    }

    return dest;
}

#ifdef __cplusplus
}
#endif

/******************************** END OF FILE ******************************/
